# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from radicale.rights import BaseRights

from .etesync_cache import EteSyncCache, etesync_for_user

import etesync as api


class Rights(BaseRights):
    def authorized(self, user, path, permission):
        if not bool(user):
            return False

        attributes = path.strip('/').split('/')

        if len(attributes) == 1:
            if attributes[0] == '':
                return permission == 'r'
            else:
                return attributes[0] == user

        if attributes[0] != user:
            return False

        if permission == 'r':
            return True

        journal_uid = attributes[1]

        with etesync_for_user(user) as (etesync, _):
            try:
                journal = etesync.get(journal_uid)
            except api.exceptions.DoesNotExist:
                return False

        return not journal.read_only



# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from http import client

import pkg_resources

from radicale import web


class Web(web.BaseWeb):
    def __init__(self, configuration, logger):
        super().__init__(configuration, logger)
        self.folder = pkg_resources.resource_filename(__name__, "web")

    def _call(self, environ, base_prefix, path, user):
        from etesync_dav.webui import app
        ret_response = []

        def start_response(status, headers):
            ret_response.append(int(status.split()[0]))
            ret_response.append(dict(headers))

        body = list(app(environ, start_response))[0]
        ret_response.append(body)
        return tuple(ret_response)

    def get(self, environ, base_prefix, path, user):
        return self._call(environ, base_prefix, path, user)

    def post(self, environ, base_prefix, path, user):
        return self._call(environ, base_prefix, path, user)



# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import base64
import json
import os


class Credentials:
    def __init__(self, filename):
        self.filename = filename
        self.last_mtime = 0
        self.content = {'users': {}}
        self.load()

    def load(self):
        if os.path.exists(self.filename):
            mtime = os.path.getmtime(self.filename)
            if mtime != self.last_mtime:
                with open(self.filename, "r") as f:
                    self.content = json.load(f)
            self.last_mtime = mtime

    def save(self):
        with open(self.filename, "w") as f:
            json.dump(self.content, f)

    def get(self, username):
        users = self.content['users']
        if username not in users:
            return None, None

        user = users[username]
        return user['authToken'], base64.b64decode(user['cipherKey'])

    def set(self, username, auth_token, cipher_key):
        users = self.content['users']
        user = {
                'authToken': auth_token,
                'cipherKey': base64.b64encode(cipher_key).decode(),
            }
        users[username] = user

    def delete(self, username):
        users = self.content['users']
        users.pop(username, None)



# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
from subprocess import check_call

from etesync_dav.config import SSL_KEY_FILE, SSL_CERT_FILE

KEY_CIPHER = 'rsa'
KEY_SIZE = 4096
KEY_DAYS = 3650  # That's 10 years.


class Error(Exception):
    pass


def has_ssl():
    return os.path.exists(SSL_KEY_FILE) and os.path.exists(SSL_CERT_FILE)


def needs_ssl():
    return sys.platform == 'darwin' and not has_ssl()


def generate_cert(cert_path: str = SSL_CERT_FILE, key_path: str = SSL_KEY_FILE,
                  key_cipher: str = KEY_CIPHER, key_size: int = KEY_SIZE,
                  key_days: int = KEY_DAYS):
    if os.path.exists(key_path):
        print('Skipping key generation as already exists.')
        return

    check_call(['openssl', 'req', '-x509', '-nodes',
                '-newkey', key_cipher + ':' + str(key_size),
                '-keyout', key_path, '-out', cert_path, '-days', str(key_days),
                '-subj', '/CN=localhost'])


def macos_trust_cert(cert_path: str = SSL_CERT_FILE, keychain: str = ''):
    if sys.platform != 'darwin':
        raise Error('this is not macOS.')
    keychain_option = ['-k', keychain] if keychain else []
    check_call(['security', 'import', cert_path] + keychain_option)
    check_call(['security', 'add-trusted-cert', '-p', 'ssl', cert_path] + keychain_option)



# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .creds import Credentials
import hashlib
import threading
import collections
import os
from contextlib import contextmanager

from urllib.parse import quote

import etesync as api
from etesync_dav import config

from .href_mapper import HrefMapper


class EteSync(api.EteSync):
    def _init_db_tables(self, database, additional_tables=[]):
        super()._init_db_tables(database, additional_tables + [HrefMapper])


class NamedReverseSemaphore:
    _data_lock = threading.RLock()
    _named_cond = collections.OrderedDict()

    def __init__(self, name):
        self.name = name

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    def _unref_name_self(self):
        cls = NamedReverseSemaphore
        with cls._data_lock:
            cls._named_cond[self.name][1] -= 1
            if cls._named_cond[self.name][1] == 0:
                del cls._named_cond[self.name]

                return True
        return False

    def acquire(self, timeout=None):
        cls = NamedReverseSemaphore
        cond = None
        with cls._data_lock:
            if not cls._named_cond:
                cls._named_cond[self.name] = [None, 1]
                return True
            elif next(iter(cls._named_cond)) == self.name:
                cls._named_cond[self.name][1] += 1
                return True
            else:
                if self.name not in cls._named_cond:
                    cond = threading.Condition()
                    cls._named_cond[self.name] = [cond, 1]
                else:
                    cond = cls._named_cond[self.name][0]
                    cls._named_cond[self.name][1] += 1
                cond.acquire()
        if cond is not None:
            ret = cond.wait(timeout=timeout)

            cond.release()
            if not ret:
                self._unref_name_self()
            return ret
        else:
            raise RuntimeError('Condition is None, should never happen!')

    def release(self):
        cls = NamedReverseSemaphore
        cond = None
        with cls._data_lock:
            owner = next(iter(cls._named_cond))
            if owner == self.name:
                if self._unref_name_self():
                    if cls._named_cond:
                        cond, refcount = next(iter(cls._named_cond.values()))
                        cond.acquire()
            else:
                raise RuntimeError('NamedReverseSemaphore: did not own lock at exit. Owner: ' + str(owner))
        if cond is not None:
            cond.notify_all()
            cond.release()

class EteSyncCache:
    def __init__(self, creds_path, db_path, remote_url=None):
        self._etesync_cache = {}
        self.creds = None
        self.creds_path = os.path.expanduser(creds_path)
        self.db_path = os.path.expanduser(db_path)
        self.remote_url = os.environ.get('ETESYNC_URL', remote_url)

    def etesync_for_user(self, user):
        if self.creds:
            # Always attempt a reload
            self.creds.load()

            # Used the cached etesync for the user unless the cipher_key or auth_token have changed.
            if user in self._etesync_cache:
                etesync = self._etesync_cache[user]
                if (etesync.auth_token, etesync.cipher_key) == self.creds.get(user):
                    etesync.reinit()
                    return etesync, False
                else:
                    del self._etesync_cache[user]
        else:
            self.creds = Credentials(self.creds_path)

        auth_token, cipher_key = self.creds.get(user)

        # Create a unique filename for user and cipher_key combos. So we don't use old caches that are no longer valid.
        unique_name_sha = hashlib.sha256(cipher_key)
        db_name_unique = '{}-{}'.format(quote(user, safe=''), unique_name_sha.hexdigest())

        db_path = self.db_path.format(db_name_unique)

        if auth_token is None:
            raise Exception('Very bad! User "{}" not found in credentials file.'.format(user))

        etesync = EteSync(user, auth_token, remote=self.remote_url, db_path=db_path)
        etesync.cipher_key = cipher_key

        self._etesync_cache[user] = etesync

        return etesync, True


_etesync_cache = EteSyncCache(
    creds_path=config.CREDS_FILE,
    db_path=config.DATABASE_FILE,
    remote_url=config.ETESYNC_URL,
)


@contextmanager
def etesync_for_user(user):
    with NamedReverseSemaphore(user):
        yield _etesync_cache.etesync_for_user(user)



# This file is part of Radicale Server - Calendar Server
# Copyright © 2011-2017 Guillaume Ayoub
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Radicale executable module.

This module can be executed from a command line with ``$python -m radicale`` or
from a python programme with ``radicale.__main__.run()``.

"""

import argparse
import atexit
import os
import select
import signal
import socket
import ssl
import sys
from wsgiref.simple_server import make_server

from radicale import (VERSION, Application, RequestHandler, ThreadedHTTPServer,
                      ThreadedHTTPSServer, config, log, storage)


class MyApplication(Application):
    def do_POST(self, environ, base_prefix, path, user):
        """Manage POST request."""
        # Dispatch .web URL to web module
        if path == "/.web" or path.startswith("/.web/"):
            return self.Web.post(environ, base_prefix, path, user)

        return super().do_POST(environ, base_prefix, path, user)


def run(passed_args=None):
    """Run Radicale as a standalone server."""
    # Get command-line arguments
    parser = argparse.ArgumentParser(usage="radicale [OPTIONS]")

    parser.add_argument("--version", action="version", version=VERSION)
    parser.add_argument("--verify-storage", action="store_true",
                        help="check the storage for errors and exit")
    parser.add_argument(
        "-C", "--config", help="use a specific configuration file")

    groups = {}
    for section, values in config.INITIAL_CONFIG.items():
        group = parser.add_argument_group(section)
        groups[group] = []
        for option, data in values.items():
            kwargs = data.copy()
            long_name = "--{0}-{1}".format(
                section, option.replace("_", "-"))
            args = kwargs.pop("aliases", [])
            args.append(long_name)
            kwargs["dest"] = "{0}_{1}".format(section, option)
            groups[group].append(kwargs["dest"])
            del kwargs["value"]
            if "internal" in kwargs:
                del kwargs["internal"]

            if kwargs["type"] == bool:
                del kwargs["type"]
                kwargs["action"] = "store_const"
                kwargs["const"] = "True"
                opposite_args = kwargs.pop("opposite", [])
                opposite_args.append("--no{0}".format(long_name[1:]))
                group.add_argument(*args, **kwargs)

                kwargs["const"] = "False"
                kwargs["help"] = "do not {0} (opposite of {1})".format(
                    kwargs["help"], long_name)
                group.add_argument(*opposite_args, **kwargs)
            else:
                group.add_argument(*args, **kwargs)

    args = parser.parse_args(passed_args)
    if args.config is not None:
        config_paths = [args.config] if args.config else []
        ignore_missing_paths = False
    else:
        config_paths = ["/etc/radicale/config",
                        os.path.expanduser("~/.config/radicale/config")]
        if "RADICALE_CONFIG" in os.environ:
            config_paths.append(os.environ["RADICALE_CONFIG"])
        ignore_missing_paths = True
    try:
        configuration = config.load(config_paths,
                                    ignore_missing_paths=ignore_missing_paths)
    except Exception as e:
        print("ERROR: Invalid configuration: %s" % e, file=sys.stderr)
        if args.logging_debug:
            raise
        exit(1)

    # Update Radicale configuration according to arguments
    for group, actions in groups.items():
        section = group.title
        for action in actions:
            value = getattr(args, action)
            if value is not None:
                configuration.set(section, action.split('_', 1)[1], value)

    if args.verify_storage:
        # Write to stderr when storage verification is requested
        configuration["logging"]["config"] = ""

    # Start logging
    filename = os.path.expanduser(configuration.get("logging", "config"))
    debug = configuration.getboolean("logging", "debug")
    try:
        logger = log.start("radicale", filename, debug)
    except Exception as e:
        print("ERROR: Failed to start logger: %s" % e, file=sys.stderr)
        if debug:
            raise
        exit(1)

    if args.verify_storage:
        logger.info("Verifying storage")
        try:
            Collection = storage.load(configuration, logger)
            with Collection.acquire_lock("r"):
                if not Collection.verify():
                    logger.error("Storage verifcation failed")
                    exit(1)
        except Exception as e:
            logger.error("An exception occurred during storage verification: "
                         "%s", e, exc_info=True)
            exit(1)
        return

    try:
        serve(configuration, logger)
    except Exception as e:
        logger.error("An exception occurred during server startup: %s", e,
                     exc_info=True)
        exit(1)


def daemonize(configuration, logger):
    """Fork and decouple if Radicale is configured as daemon."""
    # Check and create PID file in a race-free manner
    if configuration.get("server", "pid"):
        try:
            pid_path = os.path.abspath(os.path.expanduser(
                configuration.get("server", "pid")))
            pid_fd = os.open(
                pid_path, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        except OSError as e:
            raise OSError("PID file exists: %r" %
                          configuration.get("server", "pid")) from e
    pid = os.fork()
    if pid:
        # Write PID
        if configuration.get("server", "pid"):
            with os.fdopen(pid_fd, "w") as pid_file:
                pid_file.write(str(pid))
        sys.exit()
    if configuration.get("server", "pid"):
        os.close(pid_fd)

        # Register exit function
        def cleanup():
            """Remove the PID files."""
            logger.debug("Cleaning up")
            # Remove PID file
            os.unlink(pid_path)
        atexit.register(cleanup)
    # Decouple environment
    os.chdir("/")
    os.setsid()
    with open(os.devnull, "r") as null_in:
        os.dup2(null_in.fileno(), sys.stdin.fileno())
    with open(os.devnull, "w") as null_out:
        os.dup2(null_out.fileno(), sys.stdout.fileno())
        os.dup2(null_out.fileno(), sys.stderr.fileno())


def serve(configuration, logger):
    """Serve radicale from configuration."""
    logger.info("Starting Radicale")

    # Create collection servers
    servers = {}
    if configuration.getboolean("server", "ssl"):
        server_class = ThreadedHTTPSServer
        server_class.certificate = configuration.get("server", "certificate")
        server_class.key = configuration.get("server", "key")
        server_class.certificate_authority = configuration.get(
            "server", "certificate_authority")
        server_class.ciphers = configuration.get("server", "ciphers")
        server_class.protocol = getattr(
            ssl, configuration.get("server", "protocol"), ssl.PROTOCOL_SSLv23)
        # Test if the SSL files can be read
        for name in ["certificate", "key"] + (
                ["certificate_authority"]
                if server_class.certificate_authority else []):
            filename = getattr(server_class, name)
            try:
                open(filename, "r").close()
            except OSError as e:
                raise RuntimeError("Failed to read SSL %s %r: %s" %
                                   (name, filename, e)) from e
    else:
        server_class = ThreadedHTTPServer
    server_class.client_timeout = configuration.getint("server", "timeout")
    server_class.max_connections = configuration.getint(
        "server", "max_connections")
    server_class.logger = logger

    RequestHandler.logger = logger
    if not configuration.getboolean("server", "dns_lookup"):
        RequestHandler.address_string = lambda self: self.client_address[0]

    shutdown_program = False

    for host in configuration.get("server", "hosts").split(","):
        try:
            address, port = host.strip().rsplit(":", 1)
            address, port = address.strip("[] "), int(port)
        except ValueError as e:
            raise RuntimeError(
                "Failed to parse address %r: %s" % (host, e)) from e
        application = MyApplication(configuration, logger)
        try:
            server = make_server(
                address, port, application, server_class, RequestHandler)
        except OSError as e:
            raise RuntimeError(
                "Failed to start server %r: %s" % (host, e)) from e
        servers[server.socket] = server
        logger.info("Listening to %r on port %d%s",
                    server.server_name, server.server_port, " using SSL"
                    if configuration.getboolean("server", "ssl") else "")

    # Create a socket pair to notify the select syscall of program shutdown
    # This is not available in python < 3.5 on Windows
    if hasattr(socket, "socketpair"):
        shutdown_program_socket_in, shutdown_program_socket_out = (
            socket.socketpair())
    else:
        shutdown_program_socket_in, shutdown_program_socket_out = None, None

    # SIGTERM and SIGINT (aka KeyboardInterrupt) should just mark this for
    # shutdown
    def shutdown(*args):
        nonlocal shutdown_program
        if shutdown_program:
            # Ignore following signals
            return
        logger.info("Stopping Radicale")
        shutdown_program = True
        if shutdown_program_socket_in:
            shutdown_program_socket_in.sendall(b"goodbye")
    signal.signal(signal.SIGTERM, shutdown)
    signal.signal(signal.SIGINT, shutdown)

    # Main loop: wait for requests on any of the servers or program shutdown
    sockets = list(servers.keys())
    if shutdown_program_socket_out:
        # Use socket pair to get notified of program shutdown
        sockets.append(shutdown_program_socket_out)
    select_timeout = None
    if not shutdown_program_socket_out or os.name == "nt":
        # Fallback to busy waiting. (select.select blocks SIGINT on Windows.)
        select_timeout = 1.0
    if configuration.getboolean("server", "daemon"):
        daemonize(configuration, logger)
    logger.info("Radicale server ready")
    while not shutdown_program:
        try:
            rlist, _, xlist = select.select(
                sockets, [], sockets, select_timeout)
        except (KeyboardInterrupt, select.error):
            # SIGINT is handled by signal handler above
            rlist, xlist = [], []
        if xlist:
            raise RuntimeError("unhandled socket error")
        if rlist:
            server = servers.get(rlist[0])
            if server:
                server.handle_request()


if __name__ == "__main__":
    run()



# Copyright © 2017 Tom Hacohen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import pytest

import time
import uuid
import threading
from etesync_dav.radicale.etesync_cache import NamedReverseSemaphore


class ExThread(threading.Thread):
    def run(self):
        self.exc = None
        try:
            if hasattr(self, '_Thread__target'):
                # Thread uses name mangling prior to Python 3.
                self.ret = self._Thread__target(*self._Thread__args, **self._Thread__kwargs)
            else:
                self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self):
        super().join()
        if self.exc:
            raise self.exc
        return self.ret

def thread_run(name, expected, barrier=None):
    if barrier is not None:
        timeout = None
    else:
        timeout = 0
    lock2 = NamedReverseSemaphore(name)
    if barrier is not None:
        barrier.wait()
    ret = lock2.acquire(timeout)
    assert expected == ret
    if expected:
        lock2.release()


class TestNamedReverseSemaphore:
    def test_same_key_nolock(self):
        name1 = str(uuid.uuid4())

        lock1 = NamedReverseSemaphore(name1)
        lock2 = NamedReverseSemaphore(name1)
        lock3 = NamedReverseSemaphore(name1)

        lock1.acquire()
        lock2.acquire()
        lock3.acquire()

        lock1.acquire()
        lock1.acquire()

        with lock1:
            with lock2:
                with lock3:
                    with lock1:
                        pass

        lock1.release()
        lock1.release()

        lock3.release()
        lock2.release()
        lock1.release()

    def test_different_key_same_thread(self):
        name1 = str(uuid.uuid4())
        name2 = str(uuid.uuid4())

        lock1 = NamedReverseSemaphore(name1)
        lock2 = NamedReverseSemaphore(name2)

        with lock1:
            assert not lock2.acquire(0)
            assert lock1.acquire(0)
            lock1.release()

        assert lock2.acquire(0)
        lock2.release()

        with lock2:
            assert not lock1.acquire(0)
            assert lock2.acquire(0)
            lock2.release()

    def test_same_key_different_thread(self):
        name1 = str(uuid.uuid4())

        lock1 = NamedReverseSemaphore(name1)
        with lock1:
            thread = ExThread(target=thread_run, args=(name1, True), daemon=True)
            thread.start()
            thread.join()

    def test_different_key_different_thread(self):
        name1 = str(uuid.uuid4())
        name2 = str(uuid.uuid4())

        thread = ExThread(target=thread_run, args=(name2, True), daemon=True)
        thread.start()
        thread.join()

        lock1 = NamedReverseSemaphore(name1)
        with lock1:
            thread = ExThread(target=thread_run, args=(name2, False), daemon=True)
            thread.start()
            thread.join()

    def test_different_key_different_thread_wait(self):
        name1 = str(uuid.uuid4())
        name2 = str(uuid.uuid4())
        barrier = threading.Barrier(2)

        thread = ExThread(target=thread_run, args=(name2, True), daemon=True)
        thread.start()
        thread.join()

        lock1 = NamedReverseSemaphore(name1)
        with lock1:
            thread = ExThread(target=thread_run, args=(name2, True, barrier), daemon=True)
            thread.start()
            barrier.wait()
            # FIXME: hack to make sure we acquired the lock in the other thread
            time.sleep(0.2)
        thread.join()

    def test_multiple_keys_different_thread(self):
        name1 = str(uuid.uuid4())
        name2 = str(uuid.uuid4())
        name3 = str(uuid.uuid4())
        barrier = threading.Barrier(3)

        threads = []

        lock1 = NamedReverseSemaphore(name1)
        with lock1:
            threads.insert(0, ExThread(target=thread_run, args=(name2, True, barrier), daemon=True))
            threads[0].start()
            threads.insert(0, ExThread(target=thread_run, args=(name3, True, barrier), daemon=True))
            threads[0].start()

            barrier.wait()
            # FIXME: hack to make sure we acquired the lock in the other thread
            time.sleep(0.2)

        for thread in threads:
            thread.join()

    def test_multiple_keys_multiple_times_different_thread(self):
        name1 = str(uuid.uuid4())
        name2 = str(uuid.uuid4())
        name3 = str(uuid.uuid4())
        barrier = threading.Barrier(5)

        threads = []

        lock1 = NamedReverseSemaphore(name1)
        with lock1:
            threads.insert(0, ExThread(target=thread_run, args=(name2, True, barrier), daemon=True))
            threads[0].start()
            threads.insert(0, ExThread(target=thread_run, args=(name2, True, barrier), daemon=True))
            threads[0].start()
            threads.insert(0, ExThread(target=thread_run, args=(name3, True, barrier), daemon=True))
            threads[0].start()
            threads.insert(0, ExThread(target=thread_run, args=(name3, True, barrier), daemon=True))
            threads[0].start()

            barrier.wait()
            # FIXME: hack to make sure we acquired the lock in the other thread
            time.sleep(0.2)

        for thread in threads:
            thread.join()
